import datetime
import logging
import sys

import dateutil.relativedelta as rd
import numpy as np
from dateutil.parser import parse

import xlsxwriter
import xlsxwriter.utility

from edapp_python_sdk import edapppy
import pandas as pd
import sqlite3
import configparser
import os
from .sql_models import *

db_name = "edapp.db"
export_folder = "exports"


def create_export_folder():
    if not os.path.exists(f"./{export_folder}"):
        os.mkdir(export_folder)


def check_for_db():
    create_export_folder()
    path = os.path.join(export_folder, db_name)

    return sqlite3.connect(path)


def setup_sqlite(table_name, column_names, values, drop=False):
    conn = check_for_db()
    cur = conn.cursor()
    if drop:
        cur.execute(f"""DROP TABLE if exists {table_name}""")
    cur.execute(f"""CREATE TABLE if not exists {table_name} {column_names}""")
    for row in values:
        cur.execute(f"INSERT OR REPLACE INTO {table_name} VALUES {row}")
    conn.commit()
    conn.close()


def export_to_csv(df, file_name):
    if not os.path.exists(f"./{export_folder}"):
        os.mkdir(export_folder)

    df.to_csv(
        os.path.join(export_folder, f"{file_name}.csv"),
        encoding="utf-8-sig",
        index=False,
    )


def convert_to_records(df, table=None):
    if table:
        if table == "lessonprogress":
            df.insert(
                0,
                "id",
                df["userId"]
                + "_"
                + df["courseId"]
                + "_"
                + df["lessonId"]
                + "_"
                + (
                    pd.to_datetime(df["eventDateTime"]).values.astype(np.int64)
                    // 10 ** 6
                ).astype(str),
            )
        elif table == "courseprogress":
            df.insert(0, "id", df["userId"] + "_" + df["courseId"])

    # Add a column of the last time we did an export
    df["last_export"] = datetime.datetime.now()
    df = df.fillna(value="")
    records = df.to_records(index=False)

    return records


def ensure_table_exists(cur, table):
    column_names = table + "_sql"

    # Create table
    cur.execute(
        f"""CREATE TABLE if not exists {table}
                   {globals()[column_names]}"""
    )


def get_table_rows(cur, table):
    rows = cur.execute(f"SELECT * FROM {table};")
    list_of_rows = list(rows)
    if rows:
        columns = [description[0] for description in rows.description]
        return list_of_rows, columns
    else:
        return None


def df_to_dict(data, columns):
    df = to_dataframe(data, columns=columns)
    export = df_to_dictionary(df)

    return export


def database_or_export(cur, table):
    logger = logging.getLogger("ea_logger")
    ensure_table_exists(cur, table)
    rows, cols = get_table_rows(cur, table)
    if rows:
        export = df_to_dict(rows, cols)
        if table == "lessonprogress":
            possible_dates = [x["eventDateTime"] for x in export]
        else:
            possible_dates = [x["last_export"] for x in export]
        last_export = max(possible_dates)
        last_export = parse(last_export)
        if table not in [
            "surveyanswers",
            "attempts",
            "courseprogress",
            "lessonprogress"
        ]:
            one_week_ago = datetime.datetime.today() - rd.relativedelta(hours=24)
            if last_export < one_week_ago:
                print(
                    f"Table '{table}' has not been updated in the last 24 hours, updating."
                )
                logger.info(f'{table} up to date.')
                return None, None, None

        return export, cols, last_export

    else:
        return None, None, None


def export_reference_tables(ea, include_lessons):
    conn = check_for_db()
    cur = conn.cursor()

    # Users
    # Check for the existence of each table and load its data if it exists.
    users, user_cols, last_export = database_or_export(cur, "users")
    custom_fields, custom_field_cols, last_export = database_or_export(cur, "custom_fields")
    user_groups, user_group_cols, last_export = database_or_export(cur, "usergroups")
    user_child_groups, user_child_groups_cols, last_export = database_or_export(cur, "usergroups_children")
    group_users, group_user_cols, last_export = database_or_export(cur, "group_users")

    # If anything is missing, we'll go and grab the data
    if not user_groups:
        user_groups = ea.get_user_groups()
        df_groups = to_dataframe(user_groups)
        export_to_csv(df_groups, "usergroups")
        setup_sqlite(
            "usergroups", usergroups_sql, convert_to_records(df_groups), drop=True
        )
        if user_groups:
            groups_with_children = [x for x in user_groups if x["hasChildren"] is True]
            if groups_with_children:
                user_child_groups = ea.get_child_user_groups(groups_with_children)
                df_child_groups = to_dataframe(user_child_groups)
                export_to_csv(df_groups, "usergroups_children")
                setup_sqlite(
                    "usergroups_children",
                    usergroups_children_sql,
                    convert_to_records(df_child_groups),
                    drop=True,
                )

    if not users or not custom_fields or not group_users:
        users, custom_fields, group_users = ea.discover_users()
        df_users = to_dataframe(users)
        df_users = df_users.drop(columns=["customFields", "userGroups"])
        export_to_csv(df_users, "users")
        setup_sqlite("users", users_sql, convert_to_records(df_users), drop=True)

        df_custom_fields = to_dataframe(custom_fields)
        export_to_csv(df_custom_fields, "custom_fields")
        setup_sqlite(
            "custom_fields",
            custom_fields_sql,
            convert_to_records(df_custom_fields),
            drop=True,
        )
        df_group_users = to_dataframe(group_users)
        export_to_csv(df_group_users, "group_users")
        setup_sqlite(
            "group_users",
            group_users_sql,
            convert_to_records(df_group_users),
            drop=True,
        )

    # Courses
    courses, course_cols, last_export = database_or_export(cur, "courses")
    if not courses:
        courses = ea.discover_courses()
        df_courses = to_dataframe(courses)
        export_to_csv(df_courses, "courses")
        setup_sqlite("courses", course_sql, convert_to_records(df_courses), drop=True)

    # Lessons
    lessons, lesson_cols, last_export = database_or_export(cur, "lessons")
    if not lessons:
        lessons = ea.discover_lessons(courses)
        df_lessons = to_dataframe(lessons)
        export_to_csv(df_lessons, "lessons")
        setup_sqlite("lessons", lessons_sql, convert_to_records(df_lessons), drop=True)

    # Analytics
    survey_answers = export_analytics(ea, cur, "surveyanswers", "surveyanswers_sql")
    attempts = export_analytics(ea, cur, "attempts", "attempts_sql")
    courseprogress = export_analytics(ea, cur, "courseprogress", "courseprogress_sql")
    coursestatistics = export_analytics(
        ea, cur, "coursestatistics", "coursestatistics_sql"
    )
    if include_lessons is True:
        lesson_progress = export_analytics(
            ea, cur, "lessonprogress", "lessonprogress_sql"
        )
    else:
        lesson_progress = None

    return {
        "users": users,
        "user_groups": user_groups,
        "user_group_children": user_child_groups,
        "group_users": group_users,
        "custom_fields": custom_fields,
        "courses": courses,
        "lessons": lessons,
        "survey_answers": survey_answers,
        "attempts": attempts,
        "courseprogress": courseprogress,
        "coursestatistics": coursestatistics,
        "lesson_progress": lesson_progress,
    }


def export_analytics(ea, cur, to_export, sql_table):
    logger = logging.getLogger("ea_logger")
    data, cols, last_export = database_or_export(cur, to_export)
    updated_data = ea.discover_analytics(to_export, startdatetime=last_export)
    if updated_data:
        df = to_dataframe(updated_data)
        if to_export in ["coursestatistics"]:
            drop = True
        else:
            drop = False
        setup_sqlite(
            to_export,
            globals()[sql_table],
            convert_to_records(df, table=to_export),
            drop=drop,
        )
        data, cols, last_export = database_or_export(cur, to_export)
    df = to_dataframe(data)
    export_to_csv(df, to_export)

    return data


def to_dataframe(export, columns=None):
    if columns:
        return pd.DataFrame(export, columns=columns)
    else:
        return pd.DataFrame(export)


def df_to_dictionary(df):
    return df.to_dict("records")


def format_survey_data(survey_answers, users, courses, lessons):
    user_ids = []
    question_ids = []
    user_names = []
    [
        user_ids.append(x["userId"])
        for x in survey_answers
        if x["userId"] not in user_ids
    ]
    for u in user_ids:
        [user_names.append(x["email"]) for x in users if x["id"] == u]

    user_ref_dict = dict(zip(user_ids, user_names))

    [
        question_ids.append(x["questionId"])
        for x in survey_answers
        if x["questionId"] not in question_ids
    ]

    csv_header = list(survey_answers[0].keys())

    for u in user_ids:
        csv_header.append(u)

    formatted_csv = []
    for q in question_ids:
        relevant_responses = [x for x in survey_answers if x["questionId"] == q]
        new_row = relevant_responses[0]

        # Change the lesson IDs into the lesson name
        lesson_names = join_multiple_entries(relevant_responses, lessons, "lessonId")
        if lesson_names:
            new_row["lessonId"] = lesson_names

        # Create a comma separated list of the course names (useful if the course has been translated.)
        course_names = join_multiple_entries(relevant_responses, courses, "courseId")
        if course_names:
            new_row["courseId"] = course_names

        questions = join_multiple_entries(
            relevant_responses, "questions", "questionContent"
        )
        if questions:
            new_row["questionContent"] = questions

        for u in user_ids:
            user_response = [x for x in relevant_responses if x["userId"] == u]
            if user_response:
                # For this report, it's very likely that a user will have completed it more than once. As a result,
                # we'll grab the most recent response to display in the column.

                answer_dates = [x["answeredDateTime"] for x in user_response]
                most_recent_date = max(answer_dates)
                most_recent_response = [
                    x
                    for x in user_response
                    if x["answeredDateTime"] == most_recent_date
                ]
                user_response = most_recent_response[0]["answerContent"]
            else:
                user_response = None
            new_row[u] = user_response

        if new_row:
            formatted_csv.append(new_row)

    return formatted_csv, user_ref_dict


def join_multiple_entries(relevant_responses, source, col_id):
    list_of_ids = []
    [
        list_of_ids.append(x[col_id])
        for x in relevant_responses
        if x[col_id] not in list_of_ids
    ]
    names_col = []
    for item_id in list_of_ids:
        if source == "questions":
            names_col.append(item_id)
        else:
            name = [x for x in source if x["id"] == item_id]
            names_col.append(name[0]["title"])
    if len(names_col) > 1:
        names = ",\n\n".join(names_col)
    elif names_col:
        names = names_col[0]
    else:
        names = None

    return names


def surveys_to_excel(survey_data, user_ref_dict):
    df_survey_answers = pd.DataFrame(survey_data)
    df_survey_answers = df_survey_answers.rename(columns=user_ref_dict)

    df_survey_answers = df_survey_answers.drop(
        columns=[
            "id",
            "lessonExternalId",
            "courseExternalId",
            "slideId",
            "userId",
            "userExternalId",
            "attemptId",
            "questionId",
            "answerDefinitionId",
            "answerContent",
            "answeredDateTime",
            "last_export",
        ]
    )

    file_name = "survey_export"
    file_name = os.path.join(export_folder, f"{file_name}")
    workbook = xlsxwriter.Workbook("{}.xlsx".format(file_name))
    worksheet = workbook.add_worksheet("Surveys")
    worksheet.activate()

    template_as_list = df_survey_answers.values.tolist()
    columns_as_list = df_survey_answers.columns.values.tolist()

    fixed_template_as_list = []
    for values in template_as_list:
        fixed = [x.replace("\\n\\n", "\n") for x in values if x is not None]
        fixed_template_as_list.append(fixed)

    template_as_list = fixed_template_as_list

    # wrap_format = workbook.add_format({"text_wrap": 1})
    wrap_format = workbook.add_format()
    wrap_format.set_text_wrap()

    reformatted_columns = []
    for column in columns_as_list:
        current_row = {"header": column, "header_format": wrap_format}
        reformatted_columns.append(current_row)

    last_col = len(columns_as_list) - 1
    last_col_letter = xlsxwriter.utility.xl_col_to_name(last_col)

    table_length = len(template_as_list)

    worksheet.set_column(f"A1:{last_col_letter}", 25, wrap_format)
    worksheet.add_table(
        f"A1:{last_col_letter}{table_length}",
        {
            "data": template_as_list,
            "columns": reformatted_columns,
            "banded_rows": 0,
            "banded_columns": 1,
        },
    )
    workbook.close()


def write_new_config(config):
    config["TOKENS"] = {"edapp": "YOUR TOKEN HERE"}
    config["SETTINGS"] = {"start_date": "2021-01-01", "include_lesson_progress": False}
    config.write(open("config.ini", "w"))
    print(
        "Created new config file, please update with your API token and re-run the tool."
    )
    sys.exit()


def get_config():
    config = configparser.ConfigParser()
    if not os.path.isfile("config.ini"):
        write_new_config(config)
    else:
        config.read("config.ini")
        if ["TOKENS", "SETTINGS"] != config.sections():
            write_new_config(config)
        else:
            ea_api_key = config["TOKENS"]["edapp"]

    start_date = parse(config["SETTINGS"]["start_date"])
    include_lesson_progress = config["SETTINGS"]["include_lesson_progress"]

    return {
        "ea": edapppy.EdApp(ea_api_key, start_date),
        "lessons": include_lesson_progress,
    }


def export_survey_report():
    all_results = export_all()
    survey_data, user_ref_dict = format_survey_data(
        all_results["survey_answers"],
        all_results["users"],
        all_results["courses"],
        all_results["lessons"],
    )
    surveys_to_excel(survey_data, user_ref_dict)


def create_new_groups():
    ea = get_config()
    ea = ea["ea"]
    existing_groups = ea.get_user_groups()
    list_of_groups = []
    for group in list_of_groups:
        check_for_existing_group = [x for x in existing_groups if x["name"] == group]
        if check_for_existing_group:
            print("Group already exists, skipping.")
        if not check_for_existing_group:
            ea.create_group(group)


def edapp_export():
    all_results = export_all()
    if all_results:
        print('Export complete.')
    else:
        print('Nothing to export')


def export_all():
    ea = get_config()
    include_lessons = ea["lessons"]
    include_lessons = True if include_lessons.lower() == "true" else False
    ea = ea["ea"]
    # users, custom_fields, courses, lessons, survey_answers, attempts
    all_results = export_reference_tables(ea, include_lessons)

    return all_results



usergroups_sql = """(
  id text PRIMARY KEY, name text, email text, 
  hasChildren boolean, last_export datetime
)"""

group_users_sql = """(
  user_id text, group_id text, last_export datetime
)"""

users_sql = """(
  id text PRIMARY KEY, name text, email text, 
  firstName text, lastName text, externalId text, 
  lastActiveDate datetime, userGroupsManaged text, 
  roles text, last_export datetime
)"""

usergroups_children_sql = """(
    id text PRIMARY KEY, name text,
    inviteCode text, hasChildren boolean,
    parentGroupId text, last_export datetime
)"""

course_sql = """(
  id text PRIMARY KEY, externalId text, 
  title text, description text, status text, 
  locale text, duration integer, thumbnailUrl text, 
  createdDateTime datetime, modifiedDateTime datetime, 
  last_export datetime
)
"""

courses_sql = """(
  id text PRIMARY KEY, externalId text, 
  title text, description text, status text, 
  locale text, duration integer, thumbnailUrl text, 
  createDateTime datetime, modifiedDateTime text, 
  last_export datetime
)
"""

lessons_sql = """(
  id text PRIMARY KEY, externalId text, 
  title text, description text, status text, minimum_score integer, courseId text,
  last_export datetime
)
"""

lessonprogress_sql = """(
  id text PRIMARY KEY, userId text, userEmail text,
  userFirstName text, userLastName text,
  courseId text, courseTitle text,
  courseExternalId text, lessonId text,
  lessonTitle text, lessonExternalId text,
  eventName text, eventDateTime datetime,
  lessonScore integer, last_export datetime
)
"""

custom_fields_sql = """(
  user_id text, user_email text, custom_field text, 
  custom_value text, last_export datetime
)
"""

surveyanswers_sql = """(
  id text PRIMARY KEY, courseid text, courseexternalid text, 
  lessonid text, lessonexternalid text, 
  slideid text, userid text, userexternalid text, 
  attemptid text, questionid text, questioncontent text, 
  answerdefinitionid text, answercontent text, 
  answereddatetime datetime, last_export datetime
)
"""

attempts_sql = """(
  id text PRIMARY KEY, userId text, userExternalId text, 
  courseId text, courseExternalId text, 
  lessonId text, lessonExternalId text, 
  startedDateTime datetime, completedDateTime datetime, 
  score int, success boolean, earnedStars int, 
  last_export datetime
)
"""

courseprogress_sql = """(
  id text PRIMARY KEY, unlocked boolean, unlockedDateTime datetime, 
  opened boolean, openedDateTime datetime, completed boolean, 
  completedDateTime datetime, lessonsTotal integer, 
  lessonsUnlocked integer, lessonsCompleted integer, courseId text,
  courseExternalId text, courseTitle text, userName text, 
  userFirstName text, userLastName text, userId text, 
  userExternalId text, archived boolean, percentageCompleted float,
  score float, last_export datetime
)
"""

coursestatistics_sql = """(
  courseId text, totalUsers integer, 
  completedCount integer, openedCount integer, last_export datetime
)
"""



import collections
import json
import sys

import html2text
import requests
import logging
import os
import re
from datetime import datetime, timedelta, timezone
from dateutil.parser import parse
import errno
from tqdm import tqdm
import time
from random import randint
from rich.logging import RichHandler
from rich.console import Console
from rich.table import Table
from rich.progress import track

HTTP_USER_AGENT_ID = "edapp-python-sdk"


def get_log():
    logger = logging.getLogger("ea_logger")

    return logger


class EdApp:
    def __init__(self, api_token, start_date):
        self.current_dir = os.getcwd()
        self.log_dir = self.current_dir + "/log/"
        self.start_date = start_date

        # Root URLs
        self.api_url = "https://rest.edapp.com"
        self.catalog_url = self.api_url + "/v2/catalog"
        self.analytics_url = self.api_url + "/v2/analytics"

        # Users
        self.users = self.api_url + "/v2/users"
        self.user_groups = self.api_url + "/v2/usergroups"

        # Catalog
        self.course_collections = self.catalog_url + "/courseCollections"
        self.courses = self.catalog_url + "/courses"

        # Analytics
        self.survey_answers = self.analytics_url + "/surveyanswers"
        self.survey_definitions = self.analytics_url + "/surveyquestiondefinitions"
        self.attempts = self.analytics_url + "/attempts"
        self.courseProgress = self.analytics_url + "/courseprogress"
        self.courseStatistics = self.analytics_url + "/coursestatistics"
        self.lessonProgressEvents = self.analytics_url + "/lesson-progress/events"
        self.lessonStatistics = self.analytics_url + "/lessonstatistics"

        self.create_directory_if_not_exists(self.log_dir)
        self.configure_logging()
        self.api_token = api_token
        logger = logging.getLogger("ea_logger")
        if self.api_token:
            self.custom_http_headers = {
                "User-Agent": HTTP_USER_AGENT_ID,
                "Authorization": "Bearer " + self.api_token,
            }
        else:
            logger.error("No valid API token parsed! Exiting.")
            sys.exit(1)

    def authenticated_request_get(self, url):
        return requests.get(url, headers=self.custom_http_headers)

    def authenticated_request_post(self, url, data):
        self.custom_http_headers["content-type"] = "application/json"
        response = requests.post(url, data, headers=self.custom_http_headers)
        del self.custom_http_headers["content-type"]
        return response

    def authenticated_request_put(self, url, data):
        self.custom_http_headers["content-type"] = "application/json"
        response = requests.put(url, data, headers=self.custom_http_headers)
        del self.custom_http_headers["content-type"]
        return response

    def authenticated_request_delete(self, url):
        return requests.delete(url, headers=self.custom_http_headers)

    @staticmethod
    def parse_json(json_to_parse):
        """
        Parse JSON string to OrderedDict and return
        :param json_to_parse:  string representation of JSON
        :return:               OrderedDict representation of JSON
        """
        return json.JSONDecoder(object_pairs_hook=collections.OrderedDict).decode(
            json_to_parse.decode("utf-8")
        )

    @staticmethod
    def log_critical_error(ex, message):
        """
        Write exception and description message to log

        :param ex:       Exception instance to log
        :param message:  Descriptive message to describe exception
        """
        logger = logging.getLogger("ea_logger")

        if logger is not None:
            logger.critical(message)
            logger.critical(ex)

    def configure_logging(self):
        """
        Configure logging to log to std output as well as to log file
        """

        log_level = logging.DEBUG

        log_filename = datetime.now().strftime("%Y-%m-%d") + ".log"
        ea_logger = logging.getLogger("ea_logger")
        ea_logger.setLevel(log_level)
        formatter = logging.Formatter("%(asctime)s : %(levelname)s : %(message)s")

        fh = logging.FileHandler(filename=self.log_dir + log_filename)
        fh.setLevel(log_level)
        fh.setFormatter(formatter)
        ea_logger.addHandler(fh)

        sh = logging.StreamHandler(sys.stdout)
        sh.setLevel(logging.FATAL)
        sh.setFormatter(formatter)
        ea_logger.addHandler(sh)
        ea_logger.addHandler(RichHandler(level="INFO"))

    def create_directory_if_not_exists(self, path):
        """
        Creates 'path' if it does not exist

        If creation fails, an exception will be thrown

        :param path:    the path to ensure it exists
        """
        try:
            os.makedirs(path)
        except OSError as ex:
            if ex.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                self.log_critical_error(
                    ex, "An error happened trying to create " + path
                )
                raise

    def discover_analytics(
        self,
        to_export,
        courseid=None,
        courseexternalid=None,
        lessonid=None,
        lessonexternalid=None,
        userid=None,
        userexternalid=None,
        startdatetime=None,
        enddatetime=None,
        page=1,
        pagesize=1000,
    ):

        # Write out the arguments passed to the function
        func_args = locals().items()
        logger = get_log()
        # logger = logging.getLogger("ea_logger")

        result_count = None

        total_results = []

        if to_export == "surveyanswers":
            search_url = self.survey_answers
        elif to_export == "surveydefinitions":
            search_url = self.survey_definitions
        elif to_export == "attempts":
            search_url = self.attempts
        elif to_export == "courseprogress":
            search_url = self.courseProgress
        elif to_export == "coursestatistics":
            search_url = self.courseStatistics
        elif to_export == "lessonprogress":
            search_url = self.lessonProgressEvents
        else:
            logger.error(f"{to_export} is not defined.")
            sys.exit()

        search_url = search_url + "?"

        # Append any additional arguments to the URL
        for k, v in func_args:
            if v and k not in ["self", "to_export", "page", "pagesize"]:
                if k == "startdatetime" and to_export == "courseprogress":
                    k = "modifiedsincedatetime"
                logger.debug(f'Appending {k}:{v} to {search_url}')
                search_url += f"&{k}={v}"

        if to_export == "lessonprogress":
            if startdatetime:
                total_results = self.get_all_lesson_events(
                    start_date_time=startdatetime
                )
            else:
                total_results = self.get_all_lesson_events()
        else:
            total_results = self.get_all_results(search_url, to_export)

        if to_export == "surveyanswers":
            for survey in total_results:
                for k, v in survey.items():
                    if k == "questionContent":
                        cleaned_v = html2text.html2text(v)
                        survey[k] = cleaned_v.strip()

        return total_results

    def get_all_lesson_events(self, start_date_time=None):
        if not start_date_time:
            start_date_time = self.start_date
        start_date_time_str = start_date_time.strftime("%Y-%m-%dT%H:%M:%S+0")
        start_date_time = parse(start_date_time_str)
        now_plus_7 = datetime.now(timezone.utc)
        now_plus_7 = now_plus_7 + timedelta(days=7)
        all_lesson_events = []
        while start_date_time < now_plus_7:
            max_date_time = start_date_time + timedelta(days=7)
            max_date_time_str = max_date_time.strftime("%Y-%m-%dT%H:%M:%S")
            start_date_time_str = start_date_time.strftime("%Y-%m-%dT%H:%M:%S")
            current_url = f"{self.lessonProgressEvents}?&page=1&pagesize=1000&mineventdatetime={start_date_time_str}&maxeventdatetime={max_date_time_str}"
            all_lesson_events.extend(
                self.get_all_results(current_url, "lessonprogress")
            )
            start_date_time = max_date_time

        return all_lesson_events if all_lesson_events else None

    def get_all_results(
        self,
        search_url,
        to_export,
        page=1,
        pagesize=1000
    ):
        logger = logging.getLogger("ea_logger")
        result_count = None
        total_results = []
        if "?" not in search_url:
            if search_url.endswith("/"):
                search_url = search_url[:-1]
            search_url = search_url + "?"
        while result_count != 0:
            current_url = f"{search_url}&page={page}&pagesize={pagesize}"
            log_string = f"\nDownloading {to_export}: " + "\n"
            log_string += "url          = " + str(search_url) + "\n"
            log_string += "page          = " + str(page) + "\n"
            log_string += "page size     = " + str(pagesize)
            logger.debug(log_string)

            table = Table()

            table.add_column("Export", justify="center", style="cyan", no_wrap=True)
            table.add_column("URL", justify="center", style="cyan", no_wrap=True)
            table.add_column("Page", style="magenta")
            table.add_column("Page Size", justify="right", style="green")

            table.add_row(to_export, str(search_url), str(page), str(pagesize))

            console = Console()
            console.print(table)

            response = self.authenticated_request_get(current_url)
            result = (
                response.json() if response.status_code == requests.codes.ok else None
            )
            number_discovered = str(result["totalCount"]) if result is not None else "0"
            log_message = (
                f"on {to_export}: "
                + number_discovered
                + " discovered using "
                + search_url
            )

            self.log_http_status(response.status_code, log_message)

            total_count = result["totalCount"]
            result = result["items"]
            result_count = len(result)
            if total_count == result_count:
                result_count = 0
            total_results.extend(result)
            page += 1

        return total_results

    def discover_lessons(self, courses):
        if courses:
            list_of_courses = self.list_dict_or_str(courses, dict_key="id")
            compiled_list = self.get_results(list_of_courses, self.courses, "lessons")
        else:
            compiled_list = None

        return compiled_list

    @staticmethod
    def list_dict_or_str(item, dict_key=None):
        if type(item) == list:
            if type(item[0]) == dict and dict_key:
                returned_list = []
                [returned_list.append(x[dict_key]) for x in item]
            else:
                returned_list = item
        else:
            returned_list = [item]

        return returned_list

    def get_results(self, list_of_items, url, item):
        logger = logging.getLogger("ea_logger")
        compiled_list = []

        # progress_bar = tqdm(list_of_items)
        for c in track(list_of_items, description="Processing..."):
            search_url = url + f"/{c}/{item}"
            logger.debug(f'Searching using {search_url}')
            log_message = f"Getting {item} from ID {c}"
            response = self.authenticated_request_get(search_url)
            self.log_http_status(response.status_code, log_message)
            result = (
                response.json() if response.status_code == requests.codes.ok else None
            )
            logger.debug(f"Found {len(result)} {item} in {c}")
            if result:
                if item == "lessons":
                    result = [dict(item, **{"courseId": c}) for item in result]
                compiled_list.extend(result)

        return compiled_list

    def get_user(self, user_id):
        search_url = self.users + f"/{user_id}"
        response = self.authenticated_request_get(search_url)
        result = response.json() if response.status_code == requests.codes.ok else None

        return result

    def get_user_groups(self):
        search_url = self.user_groups + "?"
        return self.get_all_results(search_url, "usergroups")

    def get_child_user_groups(self, group_id):
        if type(group_id) == list:
            all_child_groups = []
            for group in group_id:
                search_url = self.user_groups + "/" + group["id"] + "/children?"
                result = self.get_all_results(search_url, "children")
                add_group = [
                    dict(item, **{"parentGroupId": group["id"]}) for item in result
                ]
                all_child_groups.extend(add_group)

            return all_child_groups if all_child_groups else None
        else:
            search_url = self.user_groups + group_id + "/children?"
            result = self.get_all_results(search_url, "children")
            if result:
                return [dict(item, **{"parentGroupId": group_id}) for item in result]
            else:
                return None

    def get_users_in_group(self, group_id):
        if type(group_id) == list:
            all_group_users = []
            for group in group_id:
                search_url = self.user_groups + "/" + group["id"] + "/users?"
                all_group_users.extend(self.get_all_results(search_url, "usersgroups"))
            result = all_group_users if all_group_users else None
        else:
            search_url = self.user_groups + group_id + "/users"
            response = self.authenticated_request_get(search_url)
            result = (
                response.json() if response.status_code == requests.codes.ok else None
            )

        return result

    def discover_users(self):
        search_url = self.users + "?"
        list_of_users = self.get_all_results(search_url, to_export="users")
        custom_fields = []
        user_groups = []
        for user in list_of_users:
            user_id = user["id"]
            user_email = user["email"]
            for k, v in user.items():
                if k == "userGroups":
                    if v:
                        for y in v:
                            if "," in y:
                                y.split(",")
                                for line in y:
                                    new_row = {"user_id": user_id, "group_id": line}
                                    user_groups.append(new_row)
                            else:
                                new_row = {"user_id": user_id, "group_id": y}
                                user_groups.append(new_row)
                if k == "customFields":
                    if v != {}:
                        for x, y in v.items():
                            if y == "Hello":
                                field_title = "custom_" + x
                                new_row = {
                                    "user_id": user_id,
                                    "user_email": user_email,
                                    "custom_field": field_title,
                                    "custom_value": y,
                                }
                                custom_fields.append(new_row)
                if type(v) == list:
                    if not v:
                        user[k] = None
                    else:
                        user[k] = ",".join(v)

        return list_of_users, custom_fields, user_groups

    def discover_courses(self):
        return self.get_all_results(self.courses, to_export="courses")

    def create_group(self, group, invite_code=None):
        if not invite_code:
            invite_name = re.sub("[^A-Za-z0-9]+", "_", group)
            invite_code = invite_name + "_" + str(randint(11111, 99999))
        request = {"name": group, "inviteCode": invite_code.lower()}

        response = self.authenticated_request_post(
            self.user_groups, json.dumps(request)
        )
        result = response.json() if response.status_code == requests.codes.ok else None

        return result

    @staticmethod
    def log_http_status(status_code, message):
        """
        Write http status code and descriptive message to log

        :param status_code:  http status code to log
        :param message:      to describe where the status code was obtained
        """
        logger = get_log()
        status_description = requests.status_codes._codes[status_code][0]
        log_string = (
            str(status_code)
            + " ["
            + status_description
            + "] status received "
            + message
        )
        logger.info(log_string) if status_code == requests.codes.ok else logger.error(
            log_string
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['edapp_python_sdk']

package_data = \
{'': ['*']}

install_requires = \
['html2text>=2020.1.16,<2021.0.0',
 'openpyxl>=3.0.6,<4.0.0',
 'pandas>=1.2.3,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'rich>=10.1.0,<11.0.0',
 'tqdm>=4.58.0,<5.0.0',
 'xlsxwriter>=1.3.7,<2.0.0']

entry_points = \
{'console_scripts': ['edapp_create_groups = '
                     'edapp_python_sdk.main:create_new_groups',
                     'edapp_export = edapp_python_sdk.main:edapp_export',
                     'edapp_survey_report = '
                     'edapp_python_sdk.main:export_survey_report']}

setup_kwargs = {
    'name': 'edapp-python-sdk',
    'version': '0.1.7',
    'description': 'A Python tool for interacting with the EdApp SDK',
    'long_description': None,
    'author': 'Edd Abrahamsen-Mills',
    'author_email': 'edward.abrahamsen-mills@safetyculture.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
